@@ -19,15 +19,11 @@ Further information:
 * Code: https://github.com/scikit-hep/iminuit
 * Docs: https://iminuit.readthedocs.io
 """
+from iminuit.minuit import Minuit
+from iminuit.minimize import minimize
+from iminuit.util import describe
+from importlib import metadata
+
+__version__ = metadata.version("iminuit")
 
 __all__ = ["Minuit", "minimize", "describe", "__version__"]
-
-from .version import version as __version__
-
-try:
-    # make iminuit importable even if it is not installed yet for setup.cfg
-    from .minuit import Minuit
-    from .minimize import minimize
-    from .util import describe
-except ImportError:  # pragma: no cover
-    pass  # pragma: no cover
