@@ -1,8 +1,8 @@
 """Minuit class."""
 
 import warnings
-from . import util as mutil
-from ._core import (
+from iminuit import util as mutil
+from iminuit._core import (
     FCN,
     MnContours,
     MnHesse,
@@ -17,6 +17,7 @@ from ._core import (
 )
 import numpy as np
 import typing as _tp
+from iminuit.typing import UserBound
 
 # Better use numpy.typing.ArrayLike in the future, but this
 # requires dropping Python-3.6 support
@@ -51,11 +52,11 @@ class Minuit:
     _fmin: _tp.Optional[mutil.FMin]
     _covariance: _tp.Optional[mutil.Matrix]
 
+    # Set errordef to this for a least-squares cost function.
     LEAST_SQUARES = 1.0
-    """Set errordef to this for a least-squares cost function."""  # pragma: nocover
 
+    # Set errordef to this for a negated log-likelihood function.
     LIKELIHOOD = 0.5
-    """Set errordef to this for a negated log-likelihood function."""  # pragma: nocover
 
     @property
     def fcn(self) -> FCN:
@@ -596,10 +597,13 @@ class Minuit:
             start = np.array(args)
         del args
 
+        annotated = mutil.describe(fcn, annotations=True)
         if name is None:
-            name = mutil.describe(fcn)
+            name = list(annotated)
             if len(name) == 0 or (array_call and len(name) == 1):
                 name = tuple(f"x{i}" for i in range(len(start)))
+        elif len(name) == len(annotated):
+            annotated = {new: annotated[old] for (old, new) in zip(annotated, name)}
 
         if len(start) == 0 and len(kwds) == 0:
             raise RuntimeError(
@@ -632,6 +636,10 @@ class Minuit:
 
         self.reset()
 
+        for k, lim in annotated.items():
+            if lim is not None:
+                self.limits[k] = lim
+
     def reset(self) -> "Minuit":  # requires from __future__ import annotations
         """
         Reset minimization state to initial state.
@@ -1244,6 +1252,24 @@ class Minuit:
 
         return self
 
+    def visualize(self, plot: _tp.Callable = None):
+        """Visualize agreement of current model with data (requires matplotlib).
+
+        This generates a plot of the data/model agreement, using the current
+        parameter values, if the likelihood function supports this, otherwise
+        AttributeError is raised.
+
+        Parameters
+        ----------
+        plot : Callable, optional
+            This function tries to call the visualize method on the cost function, which
+            accepts the current model parameters as an array-like and potentially further
+            keyword arguments, and draws a visualization into the current matplotlib axes.
+            If the cost function does not provide a visualize method or if you want to
+            override it, pass the function here.
+        """
+        return self._visualize(plot)(self.values)
+
     def hesse(self, ncall: int = None) -> "Minuit":
         """
         Run Hesse algorithm to compute asymptotic errors.
@@ -1362,8 +1388,8 @@ class Minuit:
         Notes
         -----
         Asymptotically (large samples), the Minos interval has a coverage probability
-        equal to the given confidence level. The coverage probability is the probility for
-        the interval to contain the true value in repeated identical experiments.
+        equal to the given confidence level. The coverage probability is the probability
+        for the interval to contain the true value in repeated identical experiments.
 
         The interval is invariant to transformations and thus not distorted by parameter
         limits, unless the limits intersect with the confidence interval. As a
@@ -1440,7 +1466,7 @@ class Minuit:
         vname: str,
         *,
         size: int = 30,
-        bound: _tp.Union[float, _ArrayLike[float]] = 2,
+        bound: _tp.Union[float, UserBound] = 2,
         grid: _ArrayLike[float] = None,
         subtract_min: bool = False,
     ) -> _tp.Tuple[np.ndarray, np.ndarray, np.ndarray]:
@@ -1545,7 +1571,7 @@ class Minuit:
         vname: str,
         *,
         size: int = 100,
-        bound: _tp.Union[float, _ArrayLike[float]] = 2,
+        bound: _tp.Union[float, UserBound] = 2,
         grid: _ArrayLike[float] = None,
         subtract_min: bool = False,
     ) -> _tp.Tuple[np.ndarray, np.ndarray]:
@@ -1672,9 +1698,7 @@ class Minuit:
         y: str,
         *,
         size: int = 50,
-        bound: _tp.Union[
-            float, _tp.Tuple[_tp.Tuple[float, float], _tp.Tuple[float, float]]
-        ] = 2,
+        bound: _tp.Union[float, _tp.Iterable[_tp.Tuple[float, float]]] = 2,
         grid: _tp.Tuple[_ArrayLike, _ArrayLike] = None,
         subtract_min: bool = False,
     ) -> _tp.Tuple[np.ndarray, np.ndarray, np.ndarray]:
@@ -1731,7 +1755,7 @@ class Minuit:
             if xv.ndim != 1 or yv.ndim != 1:
                 raise ValueError("grid per parameter must be 1D array-like")
         else:
-            if isinstance(bound, tuple):
+            if isinstance(bound, _tp.Iterable):
                 xb, yb = bound
                 xrange = self._normalize_bound(x, xb)
                 yrange = self._normalize_bound(y, yb)
@@ -1900,11 +1924,8 @@ class Minuit:
         """
         Draw 2D Minos confidence region (requires matplotlib).
 
-        See :meth:`mncontour` for details on the interpretation of the region.
-
-        Parameters
-        ----------
-        See :meth:`mncontour`.
+        See :meth:`mncontour` for details on the interpretation of the region
+        and for the parameters accepted by this function.
 
         Examples
         --------
@@ -2090,7 +2111,7 @@ class Minuit:
 
         Parameters
         ----------
-        plot, optional : Callable or None
+        plot : Callable, optional
             To visualize the fit, interactive tries to access the visualize method on
             the cost function, which accepts the current model parameters as an array-like
             and potentially further keyword arguments, and draws a visualization into the
@@ -2108,211 +2129,210 @@ class Minuit:
         .. plot:: plots/interactive.py
             :include-source:
         """
-        try:
-            from ipywidgets import (
-                HBox,
-                VBox,
-                Output,
-                FloatSlider,
-                Button,
-                ToggleButton,
-                Layout,
-                Dropdown,
-            )
-            from IPython.display import clear_output
-            from matplotlib import pyplot as plt
-        except ModuleNotFoundError as e:
-            e.msg += (
-                "\n\nPlease install ipywidgets, IPython, and matplotlib to "
-                "enable interactive"
-            )
-            raise
-
-        pyfcn = self.fcn._fcn
-
-        if plot is None:
-            if hasattr(pyfcn, "visualize"):
-                plot = pyfcn.visualize
-            else:
-                raise ValueError(
-                    f"class {pyfcn.__class__.__name__} has no visualize method, "
-                    "please use the plot argument to pass a visualization function"
-                )
-
-        def plot_with_frame(args, from_fit, report_success):
-            trans = plt.gca().transAxes
+        with warnings.catch_warnings():
+            # ipywidgets produces deprecation warnings through use of internal APIs :(
+            warnings.simplefilter("ignore")
             try:
-                with warnings.catch_warnings():
-                    if self.fcn._array_call:
-                        plot([args], **kwargs)  # prevent unpacking of array
-                    else:
-                        plot(args, **kwargs)
-            except Exception:
-                if raise_on_exception:
-                    raise
-
-                import traceback
-
-                plt.figtext(
-                    0.01,
-                    0.5,
-                    traceback.format_exc(),
-                    ha="left",
-                    va="center",
-                    transform=trans,
-                    color="r",
+                from ipywidgets import (
+                    HBox,
+                    VBox,
+                    Output,
+                    FloatSlider,
+                    Button,
+                    ToggleButton,
+                    Layout,
+                    Dropdown,
                 )
-                return
-            if from_fit:
-                fval = self.fmin.fval
-            else:
-                fval = self.fcn(args)
-            plt.text(
-                0.05,
-                1.05,
-                f"FCN = {fval:.3f}",
-                transform=trans,
-                fontsize="x-large",
-            )
-            if from_fit and report_success:
+                from ipywidgets.widgets.interaction import show_inline_matplotlib_plots
+                from IPython.display import clear_output
+                from matplotlib import pyplot as plt
+            except ModuleNotFoundError as e:
+                e.msg += (
+                    "\n\nPlease install ipywidgets, IPython, and matplotlib to "
+                    "enable interactive"
+                )
+                raise
+
+            plot = self._visualize(plot)
+
+            def plot_with_frame(args, from_fit, report_success):
+                trans = plt.gca().transAxes
+                try:
+                    with warnings.catch_warnings():
+                        if self.fcn._array_call:
+                            plot([args], **kwargs)  # prevent unpacking of array
+                        else:
+                            plot(args, **kwargs)
+                except Exception:
+                    if raise_on_exception:
+                        raise
+
+                    import traceback
+
+                    plt.figtext(
+                        0.01,
+                        0.5,
+                        traceback.format_exc(),
+                        ha="left",
+                        va="center",
+                        transform=trans,
+                        color="r",
+                    )
+                    return
+                if from_fit:
+                    fval = self.fmin.fval
+                else:
+                    fval = self.fcn(args)
                 plt.text(
-                    0.95,
+                    0.05,
                     1.05,
-                    f"{'success' if self.valid and self.accurate else 'FAILURE'}",
+                    f"FCN = {fval:.3f}",
                     transform=trans,
                     fontsize="x-large",
-                    ha="right",
                 )
+                if from_fit and report_success:
+                    plt.text(
+                        0.95,
+                        1.05,
+                        f"{'success' if self.valid and self.accurate else 'FAILURE'}",
+                        transform=trans,
+                        fontsize="x-large",
+                        ha="right",
+                    )
 
-        class ParameterBox(HBox):
-            def __init__(self, par, val, min, max, step, fix):
-                self.par = par
-                self.slider = FloatSlider(
-                    val,
-                    min=a,
-                    max=b,
-                    step=step,
-                    description=par,
-                    continuous_update=True,
-                    layout=Layout(min_width="70%"),
-                )
-                self.fix = ToggleButton(
-                    fix, description="Fix", layout=Layout(width="3.1em")
-                )
-                self.opt = ToggleButton(
-                    False, description="Opt", layout=Layout(width="3.5em")
-                )
-                self.opt.observe(self.on_opt_toggled, "value")
-                super().__init__([self.slider, self.fix, self.opt])
+            class ParameterBox(HBox):
+                def __init__(self, par, val, min, max, step, fix):
+                    self.par = par
+                    self.slider = FloatSlider(
+                        val,
+                        min=a,
+                        max=b,
+                        step=step,
+                        description=par,
+                        continuous_update=True,
+                        layout=Layout(min_width="70%"),
+                    )
+                    self.fix = ToggleButton(
+                        fix, description="Fix", layout=Layout(width="3.1em")
+                    )
+                    self.opt = ToggleButton(
+                        False, description="Opt", layout=Layout(width="3.5em")
+                    )
+                    self.opt.observe(self.on_opt_toggled, "value")
+                    super().__init__([self.slider, self.fix, self.opt])
 
-            def on_opt_toggled(self, change):
-                self.slider.disabled = self.opt.value
-                on_slider_change(None)
+                def on_opt_toggled(self, change):
+                    self.slider.disabled = self.opt.value
+                    on_slider_change(None)
 
-        def fit():
-            if algo_choice.value == "Migrad":
-                self.migrad()
-            elif algo_choice.value == "Scipy":
-                self.scipy()
-            elif algo_choice.value == "Simplex":
-                self.simplex()
-                return False
-            else:
-                assert False  # pragma: no cover, should never happen
-            return True
+            def fit():
+                if algo_choice.value == "Migrad":
+                    self.migrad()
+                elif algo_choice.value == "Scipy":
+                    self.scipy()
+                elif algo_choice.value == "Simplex":
+                    self.simplex()
+                    return False
+                else:
+                    assert False  # pragma: no cover, should never happen
+                return True
 
-        def on_slider_change(change):
-            if out.block:
-                return
-            args = [x.slider.value for x in parameters]
-            from_fit = False
-            report_success = False
-            if any(x.opt.value for x in parameters):
-                save = self.fixed[:]
-                self.fixed = [not x.opt.value for x in parameters]
-                self.values = args
+            def on_slider_change(change):
+                if out.block:
+                    return
+                args = [x.slider.value for x in parameters]
+                from_fit = False
+                report_success = False
+                if any(x.opt.value for x in parameters):
+                    save = self.fixed[:]
+                    self.fixed = [not x.opt.value for x in parameters]
+                    self.values = args
+                    report_success = fit()
+                    args = self.values[:]
+                    out.block = True
+                    for x, val in zip(parameters, args):
+                        x.slider.value = val
+                    out.block = False
+                    self.fixed = save
+                    from_fit = True
+                with out:
+                    clear_output(wait=True)
+                    plot_with_frame(args, from_fit, report_success)
+                    with warnings.catch_warnings():
+                        warnings.simplefilter("ignore")
+                        show_inline_matplotlib_plots()
+
+            def on_fit_button_clicked(change):
+                for x in parameters:
+                    self.values[x.par] = x.slider.value
+                    self.fixed[x.par] = x.fix.value
                 report_success = fit()
-                args = self.values[:]
                 out.block = True
-                for x, val in zip(parameters, args):
+                for x in parameters:
+                    val = self.values[x.par]
+                    if val < x.slider.min:
+                        x.slider.min = val
+                    elif val > x.slider.max:
+                        x.slider.max = val
                     x.slider.value = val
                 out.block = False
-                self.fixed = save
-                from_fit = True
-            # mutil._show_inline_matplotlib_plots()
-            with out:
-                clear_output(wait=True)
-                plot_with_frame(args, from_fit, report_success)
-                mutil._show_inline_matplotlib_plots()
+                with out:
+                    clear_output(wait=True)
+                    plot_with_frame(self.values, True, report_success)
+                    show_inline_matplotlib_plots()
 
-        def on_fit_button_clicked(change):
-            for x in parameters:
-                self.values[x.par] = x.slider.value
-                self.fixed[x.par] = x.fix.value
-            report_success = fit()
-            out.block = True
-            for x in parameters:
-                val = self.values[x.par]
-                if val < x.slider.min:
-                    x.slider.min = val
-                elif val > x.slider.max:
-                    x.slider.max = val
-                x.slider.value = val
+            def on_update_button_clicked(change):
+                for x in parameters:
+                    x.slider.continuous_update = not x.slider.continuous_update
+
+            def on_reset_button_clicked(change):
+                self.reset()
+                out.block = True
+                for x in parameters:
+                    x.slider.value = self.values[x.par]
+                out.block = False
+                on_slider_change(None)
+
+            parameters = []
+            for par in self.parameters:
+                val = self.values[par]
+                step = mutil._guess_initial_step(val)
+                a, b = self.limits[par]
+                # safety margin to avoid overflow warnings
+                a = a + 1e-300 if np.isfinite(a) else val - 100 * step
+                b = b - 1e-300 if np.isfinite(b) else val + 100 * step
+                parameters.append(ParameterBox(par, val, a, b, step, self.fixed[par]))
+
+            fit_button = Button(description="Fit")
+            fit_button.on_click(on_fit_button_clicked)
+
+            update_button = ToggleButton(True, description="Continuous")
+            update_button.observe(on_update_button_clicked)
+
+            reset_button = Button(description="Reset")
+            reset_button.on_click(on_reset_button_clicked)
+
+            algo_choice = Dropdown(
+                options=["Migrad", "Scipy", "Simplex"], value="Migrad"
+            )
+
+            ui = VBox(
+                [
+                    HBox([fit_button, update_button, reset_button, algo_choice]),
+                    VBox(parameters),
+                ]
+            )
+
+            out = Output()
             out.block = False
-            with out:
-                clear_output(wait=True)
-                plot_with_frame(self.values, True, report_success)
-                mutil._show_inline_matplotlib_plots()
 
-        def on_update_button_clicked(change):
             for x in parameters:
-                x.slider.continuous_update = not x.slider.continuous_update
+                x.slider.observe(on_slider_change, "value")
 
-        def on_reset_button_clicked(change):
-            self.reset()
-            out.block = True
-            for x in parameters:
-                x.slider.value = self.values[x.par]
-            out.block = False
+            show_inline_matplotlib_plots()
             on_slider_change(None)
 
-        parameters = []
-        for par in self.parameters:
-            val = self.values[par]
-            step = mutil._guess_initial_step(val)
-            a, b = self.limits[par]
-            # safety margin to avoid overflow warnings
-            a = a + 1e-300 if np.isfinite(a) else val - 100 * step
-            b = b - 1e-300 if np.isfinite(b) else val + 100 * step
-            parameters.append(ParameterBox(par, val, a, b, step, self.fixed[par]))
-
-        fit_button = Button(description="Fit")
-        fit_button.on_click(on_fit_button_clicked)
-
-        update_button = ToggleButton(True, description="Continuous")
-        update_button.observe(on_update_button_clicked)
-
-        reset_button = Button(description="Reset")
-        reset_button.on_click(on_reset_button_clicked)
-
-        algo_choice = Dropdown(options=["Migrad", "Scipy", "Simplex"], value="Migrad")
-
-        ui = VBox(
-            [
-                HBox([fit_button, update_button, reset_button, algo_choice]),
-                VBox(parameters),
-            ]
-        )
-
-        out = Output()
-        out.block = False
-
-        for x in parameters:
-            x.slider.observe(on_slider_change, "value")
-        # mutil._show_inline_matplotlib_plots()
-        on_slider_change(None)
-
-        return HBox([out, ui])
+            return HBox([out, ui])
 
     def _free_parameters(self) -> _tp.Set[str]:
         return set(mp.name for mp in self._last_state if not mp.is_fixed)
@@ -2324,7 +2344,7 @@ class Minuit:
         return pr
 
     def _normalize_bound(
-        self, vname: str, bound: _tp.Union[float, _ArrayLike[float]]
+        self, vname: str, bound: _tp.Union[float, UserBound, _tp.Tuple[float, float]]
     ) -> _tp.Tuple[float, float]:
         if isinstance(bound, _tp.Iterable):
             return mutil._normalize_limit(bound)
@@ -2428,6 +2448,11 @@ class Minuit:
             s += self.merrors._repr_html_()
         if self.covariance is not None:
             s += self.covariance._repr_html_()
+        if self.fmin is not None:
+            try:
+                s += self.visualize()._repr_html_()
+            except (AttributeError, ValueError):
+                pass
         return s
 
     def _repr_pretty_(self, p, cycle):
@@ -2436,6 +2461,18 @@ class Minuit:
         else:
             p.text(str(self))
 
+    def _visualize(self, plot):
+        pyfcn = self.fcn._fcn
+        if plot is None:
+            if hasattr(pyfcn, "visualize"):
+                plot = pyfcn.visualize
+            else:
+                raise AttributeError(
+                    f"class {pyfcn.__class__.__name__} has no visualize method, "
+                    "please use the plot argument to pass a visualization function"
+                )
+        return plot
+
 
 def _make_init_state(
     pos2var: _tp.Tuple[str, ...], args: np.ndarray, kwds: _tp.Dict[str, float]
